// MAX GONG
// POOM
// 8-28-20



/*
 * FOR USERS BESIDES ME (THE CREATOR):
 *   ctrl-F and find "ADJUST THIS" to adjust to your liking
 *   
 * LIBRARIES:
 *   FastLED.h
 *   fix_fft.h
 * 
 * IMPORTANT COMMENTS:
 * 1. Sampling rate
 * Sampling rate is ~37kHz by setting ADC prescale to 32
 * For a N=256 DFT, this means frequency bins produced by fix_fft.h have a bandwidth of 37kHz / 256 = ~144Hz
 * Explanation: 
 * Sampling rate = (ADC clock speed of 16MHz / ADC prescale factor) / 13, 
 * This is set by ADCSRA, ADPS#, and ADC conversion takes 13 ADC clocks
 * My code does this in 1) SAMPLING RATE defines under DEFINES, and 2) void setup by setting prescale to 32
 * (In the code, sampling rate is separate from Baud rate.
 * Sampling 256 samples is within its own for loop, so any 'Serial.println' statements does not induce any delay.)
 * 
 * 2. Choosing frequency bins
 * I condense N=256 frequency bins with bandwidth 144Hz from data_avg[256] 
 * into 3 general frequency bins (Low, Mid, High) in value[3]:
 *   Low - deep bass          (bins #1  -  0-144 Hz)
 *   Mid - mild bass + kicks  (bins #2-3  -  144-432 Hz)
 *   High - everything else   (bins #3-30  -  432-4320 Hz)
 * I ended up using only Mid and High to create lighting effects.
 * Individual gains multiply with value[3] to control levels.  
 * This is because High bin was always low due to less dB at higher frequencies in typical music
 * yet averaginng of 28 bins. So gain of High is more than gain of Mid, for example. 
 * (These frequencies were determined by personal preference on lighting effects and how speakers sound in my room.)
 * (I also wanted my Mid to react to kicks, so I tried calculating rising edge. However, this created major delays in lighting effect response.)
 * 
 * 3. Lighting effects & patterns
 * 
 * 
*/


#define FASTLED_ALLOW_INTERRUPTS 0
#include <FastLED.h>
#include <fix_fft.h>


////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES /////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

// PIN defines
#define AUDIO_PIN       A5 
#define POT_PIN         A1
#define LED_PIN         3
#define LED_PIN2        2
#define LED_PIN3        6

// FIX_FFT defines
#define NFFT            256
#define NFFTPW2         8                 // 7 if Nfft = 128, 8 if Nfft = 256
#define MAXNUMREADINGS  20
int x = 0;
char data[NFFT], im[NFFT];          // data is Real input & output of FFT, im is Imaginary output of FFT
char data_avgs[NFFT];               // 
float value[3];                     // # of frequency bins (3 = low, mid, high)
float levelLOW, levelMID, levelHIGH;

// Moving average filter defines
int numreadings[3];              
float values_movavg[3][MAXNUMREADINGS];
float total[3];
int readIndex[3];

// FASTLED defines
#define NUM_LEDS          100
#define MAX_BRIGHTNESS    250
#define FRAMES_PER_SECOND 120
CRGB leds[NUM_LEDS];

// LED pattern defines
uint8_t PatternNumberMID = 0; 
uint8_t PatternNumberHIGH = 0;
uint8_t PatternNumberHIGHONLY = 0;
uint8_t myCurrentModeNumber = 0;
typedef void (*SimplePatternList[])();
int i=0,j=0,audioVal;
float brightness, brightnessHIGH;
unsigned long fadeTimer, fadeTimer1;
float ihue = 0.00;
#define ARRAY_SIZE(A) (sizeof(A) / sizeof((A)[0]))


////////////////////////////////////////////////////////////////////////////////////////////////
// SETUP ///////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

void setup() {

  // Set prescale to 32 to increase sampling rate to ~37kHz
  ADCSRA = (1<<ADEN)|(1<<ADPS2)|(0<<ADPS1)|(1<<ADPS0);          // (1 0 1 corresponds to division by 32)

  // Set up LEDs
  FastLED.addLeds<WS2812, LED_PIN, GRB>(leds, NUM_LEDS);
  FastLED.addLeds<WS2812, LED_PIN2, GRB>(leds, NUM_LEDS);
  FastLED.addLeds<WS2812, LED_PIN3, GRB>(leds, NUM_LEDS);
  FastLED.setBrightness(MAX_BRIGHTNESS);

  // Define input pins 
  pinMode(AUDIO_PIN, INPUT);
  pinMode(POT_PIN, INPUT);

  // Begin serial & bluetooth communication
  Serial.begin(9600);
  
  // Set number of readings in moving average filter window
  numreadings[0] = 3; numreadings[1] = 3; numreadings[2] = MAXNUMREADINGS; 
  
  // Set initial moving average values to 0
  for (j=0; j<3; j++){
    for (i=0; i<numreadings[j]; i++){ 
      values_movavg[j][i] = 0; 
    }   
  } 
  for(i=0; i<3; i++){
    total[i] = 0; 
    readIndex[i] = 0;      
  }
}


// LED patterns & modes
SimplePatternList myPatternsMID = {             // LED patterns for MID frequency bin 
  appearandfadeMG,
  appearSegmentsMG,
  sineMG,
  appearRainbow
};
SimplePatternList myPatternsHIGH = {            // LED patterns for HIGH frequency bin
  randomGlitter,
  rainbowGlitter
};
SimplePatternList myPatternsHIGHONLY = {        // LED patterns for HIGH frequency bin
  cometNstars
};
SimplePatternList MASTERModes = {               // LED patterns in general
  reactToMusic,
  appearRainbowNORMAL,
  cometNstarsNORMAL,
  rainbowGlitterNORMAL,
  turnOFF
};
const int num_MASTERModes = ARRAY_SIZE(MASTERModes);  // number of MASTERModes 
int potentBins[num_MASTERModes];                // initialize array of potentiometer borders to help select MASTERMode
int startingState = 1;                          // Need this to define potentiometer bins, since I can't define it in setup()

// Pattern variables 
int randnumMID = 1;                             // random number variable for MID pattern LED location
int segmentLength = 40;                         // segment length for comet pattern
double posRot = -double(segmentLength);         // position of LED to increment 

// FFT bin parameters
int low_mid_threshold_bin = 1;                  // Low - deep bass (bin #1 - 0-144 Hz)
int mid_high_threshold_bin = 2;                 // Mid - mild bass + kicks (bin #2-3 - 144-432 Hz)
int high_threshold_bin = 30;                    // High - everything else (bin #3-30 - 432-4320 Hz)
float low_gain = 1;                             // Gain multipled with LOW 
float mid_gain = 2;                             // Gain multiplied with MID
float high_gain = 25;                           // Gain multiplied with HIGH
float baselineLOW = 10.0;                       // Set lower for more sensitivity and higher for less.
float maxTriggerLOW = 40.0;                     // Any volume above this saturates to MAX_BRIGHTNESS
float baselineMID = 28.0;                       // custom adjust TRIAL AND ERROR
float maxTriggerMID = baselineMID+5;           // custom adjust TRIAL AND ERROR
float baselineHIGH = 24;                        // custom adjust TRIAL AND ERROR
float maxTriggerHIGH = baselineHIGH+20;

// Hue + gradual changes + timing
float hueStep = 0.25;                           // rate at which hue changes
float fadeAmountMID = 27;                       // control fade amount for MID          
float fadeAmountHIGH = 30;                      // control fade amount for HIGH
float fadeAmountHIGHONLY = 9;                   // control fade amount for HIGHONLY
unsigned long highONLYtimer = millis();         // helps determine if high only state after some time
unsigned long time_diff = 0;                    // helps determine if high only state after some time
int highONLYstate = 0;                          // determines if high only state


////////////////////////////////////////////////////////////////////////////////////////////////
// LOOP ////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

void loop(){  
  readPotentiometer();                                // see if anything received in Bluetooth & act on it
  hue(hueStep);                                   // increment hue
  MASTERModes[myCurrentModeNumber]();             // run the MASTER Mode selected
  FastLED.show();                               // display LEDs (send to data lines)
  FastLED.delay(1000/FRAMES_PER_SECOND);        // achieve desired frame rate
}



////////////////////////////////////////////////////////////////////////////////////////////////
// CONTROL LIGHTING ////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

// put in OFF state
void turnOFF(){
  FastLED.clear();
}


// Potentiometer function to change modes/patterns
void readPotentiometer(){
  if (startingState == 1){                                 // if initial starting state, define potentiometer bins
    int bin_size = 1023 / num_MASTERModes;          // Define potentiometer "bins" depending on ARRAY_SIZE(MASTERModes)
    for (int i = 0; i < num_MASTERModes - 1; i++){
      potentBins[i] = (i+1)*bin_size;
      Serial.println(potentBins[i]);
    }
    potentBins[num_MASTERModes-1] = 1023;
    startingState = 0;
  }
  int potValue = analogRead(POT_PIN);
  for (int i = 0; i < num_MASTERModes; i++){          // choose the mode based on potentiometer value
    if (potValue <= potentBins[i]){
      myCurrentModeNumber = i;
      break;
    }
  }
  Serial.println(myCurrentModeNumber);
}

// MASTER FUNCTION to react to music
void reactToMusic(){

  EVERY_N_SECONDS( 16 ){                       // rotate through patterns every N seconds
    PatternNumberMID = (PatternNumberMID + 1) % ARRAY_SIZE(myPatternsMID);
  }

  EVERY_N_SECONDS( 10 ){                       // rotate through patterns every N seconds
    PatternNumberHIGH = (PatternNumberHIGH + 1) % ARRAY_SIZE(myPatternsHIGH);
    posRot = -double(segmentLength);
    if (highONLYstate == 1){
      PatternNumberHIGHONLY = (PatternNumberHIGHONLY + 1) % ARRAY_SIZE(myPatternsHIGHONLY);
    }
  }
  
  getFreqBins();                               // get frequency bins using FFT (3 freq bins: low, mid, high)
  levelLOW = value[0];                                  
  levelMID = value[1];
  levelHIGH = value[2];          
  
  triggerLEDforMID();                          // trigger LEDs for MID frequency bin + control fading
  triggerLEDforHIGH();                         // trigger LEDs for HIGH frequency bin + control fading
  viewFreqBins();                              // print statements to view frequency bins + brightness
}


// Trigger LEDs for MID frequncy bin + control fading 
void triggerLEDforMID(){
  if (levelMID > baselineMID){                 // trigger kick lighting, if MID frequency bin level above threshold
    brightness = min((levelMID/maxTriggerMID)*MAX_BRIGHTNESS, MAX_BRIGHTNESS);        // set brightness    
    myPatternsMID[PatternNumberMID]();                                                // run lighting pattern for MID 
    time_diff = 0;
  }
  else if (brightness > 0) {                      // control fading (need if-loop to not affect fading of HIGHs)
    fadeToBlackBy(leds, NUM_LEDS, fadeAmountMID); 
    brightness = max(brightness - fadeAmountMID/5, 0);     // have to divide because there's a lag with fadeToBlackBy
    highONLYtimer = millis();                    // start the highs-only timer
  }
  else {
    randnumMID = random(20);        // generate random number for segment location, 1-20 positions 
  }
}


// Trigger LEDs for HIGH frequency bin + control fading
void triggerLEDforHIGH(){
  if (levelHIGH > baselineHIGH) {             // trigger HIGH lighting, if HIGH frequency bin level above threshold
    brightnessHIGH = min((levelHIGH/maxTriggerHIGH)*250, 250);   //  where range of brightnessHIGH is 0-200 (out of 255)
    if (levelMID < baselineMID && time_diff > 5000) {            // greater than 5 seconds since no MID present
      highONLYstate = 1;                                         // define state for switching pattern number
      myPatternsHIGHONLY[PatternNumberHIGH]();
    }
    else {
      time_diff = millis() - highONLYtimer;
      highONLYstate = 0;
      myPatternsHIGH[PatternNumberHIGH]();                             // run lighting pattern for HIGH
    }
  }
  if (highONLYstate == 1){
    fadeToBlackBy(leds, NUM_LEDS, fadeAmountHIGHONLY);      
  }
  else {
    fadeToBlackBy(leds, NUM_LEDS, fadeAmountHIGH);      // control fading (need if-loop to not affect fading of MIDs) 
  }
}



////////////////////////////////////////////////////////////////////////////////////////////////
// CUSTOM PATTERNS /////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////


// custom sine function (say if LEDs were to react to bass) FOR MUSIC
void sineMG(){
  CRGBPalette16 palette = RainbowColors_p;
  uint8_t beat = beatsin8(150, 64, 255);                       // sine wave (BPM, low brightness, high brightness)
  for( int i = 1; i < NUM_LEDS; i+=2) {                                      
    leds[i] = ColorFromPalette(palette, ihue+(i), beat-ihue+(i*5));    // ColorfromPallete( the pallete, index in pallete, brightness )
  }
}


// custom appear and fade (MID)
void appearandfadeMG(){
  for(int led = 1; led < NUM_LEDS; led+=2) { 
    leds[led] = CHSV( ihue, 255, brightness ); 
  }
}


// custom appear and fade segments (MID)
void appearSegmentsMG(){    
  for(int led = randnumMID; led < randnumMID + 20; led++) {       // light up 20 LED segment at random location
      leds[led] = CHSV( ihue, 255, brightness ); 
      leds[led+40] = CHSV( ihue, 255, brightness ); 
  }
}


// rainbow (FastLED has own rainbow function) - FOR MUSIC
void appearRainbow() {
  for(int led = 1; led < NUM_LEDS; led+=2){
    int temphue = ihue-led;
    if (temphue  > 255) { temphue = temphue - 256; }  
    leds[led] = CHSV( temphue, 255, brightness ); 
  }
}


// rainbow (FastLED has own rainbow function) NORMAL
void appearRainbowNORMAL() {
  ihue++;
  fill_rainbow(leds, NUM_LEDS, ihue, 5);
  fadeToBlackBy(leds, NUM_LEDS, 30);
}

// glitter
void randomGlitter() {
  int temphue = ihue+50;
  if (temphue > 255) { temphue = temphue - 256; }
  if( random8() < 120) {                                // between 0-255 (since using random8())
    for (int i = 0; i < 4; i++){                        // choose i < n, n LEDs to randomly glitter                          
      leds[random16(NUM_LEDS/2)*2] = CHSV ( temphue, 255, brightnessHIGH );
    }
  }
}


// rainbow glitter FOR MUSIC
void rainbowGlitter() {
  if( random8() < 120) {                                // between 0-255 (since using random8())
    for (int i = 0; i < 4; i++){                        // choose i < n, n LEDs to randomly glitter                          
      leds[random16(NUM_LEDS/2)*2] = CHSV ( random8(), 255, brightnessHIGH );     // use a random hue
    }
  }
}


// rainbow glitter NORMAL
void rainbowGlitterNORMAL() {
  if( random8() < 100) {                                // between 0-255 (since using random8())
    for (int i = 0; i < 2; i++){                        // choose i < n, n LEDs to randomly glitter                          
      leds[random16(NUM_LEDS)] = CHSV ( random8(), 255, MAX_BRIGHTNESS-50 );     // use a random hue
    }
  }
  fadeToBlackBy(leds, NUM_LEDS, 5);
}


// comet and glitter
void cometNstars(){
  
  // comet
  rotatePos(double(segmentLength));
  int tempDivisor = (brightnessHIGH/(segmentLength/2));
  int tempPos = int(posRot);
  for(int i = 0; i <= segmentLength/2; i++){                             // fade up the beginning of segment
    if (tempPos+i >= 0) {
      leds[tempPos+i] = CHSV(ihue - i, 255, brightnessHIGH - (segmentLength/2 - i) * tempDivisor); 
    }
  }
  for(int i = segmentLength/2+1; i < segmentLength; i++){               // fade down end of segment
    if (tempPos+i >= 0) {
      leds[tempPos+i] = CHSV(ihue - i, 255, brightnessHIGH - (i - segmentLength/2) * tempDivisor); 
    }
  }

  // stars   (basically randomGlitterSLOW())
  if(random8() < 95) {
    for (int i = 0; i < 2; i++) {
      int tempIndex = random16(NUM_LEDS);
      if (tempIndex < tempPos || tempIndex > tempPos + segmentLength) {                      // so it doesn't shine on the lit up segment
        leds[tempIndex] = CHSV (ihue + 50 + int(random8()/255 * 30), 255, brightnessHIGH);
      }
    }
  }
}



// comet and glitter
void cometNstarsNORMAL(){
  
  posRot-=0.25;
  
  // comet
  rotatePos(double(segmentLength));
  int tempDivisor = (MAX_BRIGHTNESS/(segmentLength/2));
  int tempPos = int(posRot);
  for(int i = 0; i <= segmentLength/2; i++){                             // fade up the beginning of segment
    if (tempPos+i >= 0) {
      leds[tempPos+i] = CHSV(ihue - i, 255, MAX_BRIGHTNESS- (segmentLength/2 - i) * tempDivisor); 
    }
  }
  for(int i = segmentLength/2+1; i < segmentLength; i++){               // fade down end of segment
    if (tempPos+i >= 0) {
      leds[tempPos+i] = CHSV(ihue - i, 255, MAX_BRIGHTNESS - (i - segmentLength/2) * tempDivisor); 
    }
  }

  // stars   (basically randomGlitterSLOW())
  if(random8() < 95) {
    for (int i = 0; i < 1; i++) {
      int tempIndex = random16(NUM_LEDS);
      if (tempIndex < tempPos || tempIndex > tempPos + segmentLength) {                      // so it doesn't shine on the lit up segment
        leds[tempIndex] = CHSV (ihue + 50 + int(random8()/255 * 50), 255, MAX_BRIGHTNESS);
      }
    }
  }

  fadeToBlackBy(leds, NUM_LEDS, 25);
}






////////////////////////////////////////////////////////////////////////////////////////////////
// SHOWING FREQUENCY BINS  /////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////


// MAIN FUNCTION to run FFT and produce LOW|MID|HIGH in value[3]
void getFreqBins(){
  record();
  fix_fft(data, im, NFFTPW2, 0);                   // data is real, im is imaginary, 2**7 = 128, 0 for FFT (1 for iFFT)                                        
  getMagnitudeFreqBins();
  generateLowMidHigh();
  moving_average_filter(); 
}


// collect N = 256 samples for FFT (max is 256) from microphone
void record() {
  for (i=0; i < NFFT; i++){                                                          
    audioVal = analogRead(AUDIO_PIN);                                  
    data[i] = audioVal;                                     
    im[i] = 0;                                                   
  }
}

// from NFFT/2 frequency bins, calculate magnitude from real & imaginary parts
void getMagnitudeFreqBins(){
  for (i=0; i< NFFT/2; i++){                                               
    data_avgs[i] = sqrt(data[i] * data[i] + im[i] * im[i]); 
  } 
}


// Generate bins for low, mid, and high (udpate value[i])
void generateLowMidHigh() {
  value[0] = 0;
  value[1] = 0;
  value[2] = 0;
  
  // condense NFFT/2 bins to 3 bins
  for (i=0; i<low_mid_threshold_bin; i++)                       { value[0] += data_avgs[i]; }
  for (i=low_mid_threshold_bin; i<mid_high_threshold_bin; i++)  { value[1] += data_avgs[i]; }
  for (i=mid_high_threshold_bin; i<NFFT/2; i++)                 { value[2] += data_avgs[i]; }
  
  // averaging
  if (low_mid_threshold_bin - 1 != 0)                       {value[0] = value[0] / (low_mid_threshold_bin - 1);}
  if (mid_high_threshold_bin - low_mid_threshold_bin != 0)  {value[1] = value[1] / (mid_high_threshold_bin- low_mid_threshold_bin);}
  if (NFFT/2 - mid_high_threshold_bin != 0)                 {value[2] = value[2] / (NFFT/2 - mid_high_threshold_bin);}
  
  // apply gains 
  value[0] = value[0] * low_gain;
  value[1] = value[1] * mid_gain;
  value[2] = value[2] * high_gain;
}


// Perform moving average for all value[i]
void moving_average_filter() {
  for (i=0; i<3; i++) {
    total[i] = total[i] - values_movavg[i][readIndex[i]];      // subtract earliest value from total
    values_movavg[i][readIndex[i]] = value[i];                 // replace earliest value with newest value 
    total[i] = total[i] + values_movavg[i][readIndex[i]];      // add newest value to total
    readIndex[i] = readIndex[i] + 1;                // increment readIndex
    if (readIndex[i] >= numreadings[i]){ readIndex[i] = 0; }
    value[i] = total[i] / numreadings[i];             // update values[i]
  }
}



////////////////////////////////////////////////////////////////////////////////////////////////
// OTHER CUSTOM FUNCTIONS //////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

// Rotate hue
void hue(float istep){
  ihue = ihue + istep;
  if (ihue > 255) {ihue=ihue-256;}
}


// Rotate position value 
void rotatePos(double segmentL){
  posRot+=0.5;
  if (posRot >= NUM_LEDS + segmentL/2) { posRot = -segmentL; } 
}



// View frequency bins 
void viewFreqBins() {
  Serial.print(brightness);
  Serial.print(", ");
  Serial.print(brightnessHIGH);
  Serial.print(", ");
  Serial.print(value[1]);
  Serial.print(", ");
  Serial.print(value[2]);
  Serial.println();
}
